from __future__ import print_function, division
import collections
import functools

# this package creates a memoized decorator for functions
allmemos = []

def reset():
    for memo in allmemos:
        memo.reset()

class memoized(object):
    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''
    def __init__(self, func):
        self.func = func
        self.cache = {}
        allmemos.append(self)

    def reset(self):
        del self.cache
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.abc.Hashable):
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)



from __future__ import print_function, division
import collections
import itertools
import sys
from time import time

# wraps any iterable, and prints a progress report as it runs through
# that iterable
def show_progress(iterable, increment_percent=1, verbose=False, total=100):
    # if we can determine the size of the iterable, then use that to determine progress
    if isinstance(iterable, collections.Sized):
        total = len(iterable)
    stepsize = max(int(increment_percent / 100 * total), 1)
    t0 = time()
    for i, x in itertools.izip(xrange(total), iterable):
        if i % stepsize == 0:
            if not verbose:
                print('.', end='')
            else:
                print(
                    int((time() - t0) * 100) / 100,
                    ':',
                    int(i / total * 10000) / 100,
                    'percent complete (' + str(i) + ' of ' + str(total) + ')')
            sys.stdout.flush()
        yield x
    print('\n')

# returns chunks of a certain size
def grouper(chunk_size, iterable):
    it = iter(iterable)
    while True:
        chunk = list(itertools.islice(it, chunk_size))
        if not chunk:
            return
        yield chunk



from __future__ import print_function, division

# print a namesapce (as is given by, e.g, argparse) with nice formatting
def print_namespace(ns, name_width=25):
    for name, val in sorted(vars(ns).items()):
        if not name.startswith('_'):
            print(('{:<'+str(name_width)+'}{}').format(name, val))



from __future__ import print_function, division
import os, resource

# create a directory if it doesn't exist already
def makedir(path_to_dir):
    if not os.path.exists(path_to_dir):
        os.makedirs(path_to_dir)

# create a directory to house a given file
def makedir_for_file(path_to_file):
    makedir(os.path.dirname(path_to_file))

# add a prefix to a given filename
def add_prefix(path_to_file, prefix):
    directory, fname = os.path.split(path_to_file)
    if not directory:
        directory = '.'
    if fname:
        return '/'.join([directory, prefix + fname])
    else:
        print('WARNING: cannot add prefix to directory', path_to_file)
        return path_to_file

# return the filename corresponding to a hidden version of a given file
def make_hidden(path_to_file):
    return add_prefix(path_to_file, '.')

# replace a file's extension
def replace_ext(path_to_file, new_extension):
    root, ext = os.path.splitext(path_to_file)
    return root + '.' + new_extension

# return memory usage in MB
def mem():
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1000




from __future__ import print_function, division
import subprocess
import math
import pretty

# produces an LSF-formatted bsub command
def bsub_command(command, outfilepath, jobname=None, queue='short', time_in_hours=12,
        time_in_minutes=None, memory_GB=8, depends_on=None):
    if time_in_minutes:
        time = '0:' + str(time_in_minutes)
    else:
        time = str(time_in_hours) + ':00'
    # the commented line below is to filter out ottavinos, but those should usually be fast
    return ['bsub'] + \
        (['-J', jobname] if jobname else []) + \
        ['-q', queue,
        '-W', time,
        '-oo', outfilepath,
        '-R', '"rusage[mem=' + str(int(memory_GB * 1024)) + \
                # '] select[model!=XeonE5345 && model!=XeonE5430 && model!=XeonE52680]"'] + \
                '] select[model!=XeonE5345 && model!=XeonE5430]"'] + \
            (['-w', 'done({})'.format(depends_on)] if depends_on is not None else []) + \
        [' '.join(command)]

# submits a command to LSF
def submit(command, outfilepath, jobname=None, queue='short', time_in_hours=12,
        time_in_minutes=None, memory_GB=8, depends_on=None, debug=False):
    bsub_cmd = bsub_command(command, outfilepath, jobname, queue,
            time_in_hours, time_in_minutes, memory_GB, depends_on)
    print('\033[92m' + ' '.join(bsub_cmd) + '\033[0m')
    print('\033[94m' + outfilepath + '\033[0m')
    if not debug:
        output = subprocess.check_output(bsub_cmd)
        print(output)
        return output.replace('[','<').replace('>','<').split('<')[1] # parse out job id
    else:
        return 'JOBID'

# retrieves the elments of some array that are meant to be processed in a given batch
# batch_num is 1-indexed, as with LSF
def elements_in_batch(elements, batch_num, num_per_batch):
    return elements[(batch_num - 1) * num_per_batch:
            min(batch_num * num_per_batch, len(elements))]

# determines how many batches will be required to process a given number of elements
def num_batches(num_elements, num_per_batch):
    return int(math.ceil(num_elements / num_per_batch))

# creates main, submit, and merge argparsers, with each set to be triggered by the
# corresponding command-line keyword. this is for instances in which we want all three
# functions in the same file, with the ability to specify via command-line which one we
# will run. See choose_subparser_and_run
def add_main_and_submit(parser, main_function, submit_function,
        merge_function=None,
        main_parents=[],
        submit_parents=[],
        merge_parents=[]):
    subparsers = parser.add_subparsers()

    subparser_main = subparsers.add_parser('main', parents=main_parents)
    subparser_main.set_defaults(_func=main_function)

    subparser_submit = subparsers.add_parser('submit', parents=submit_parents)
    subparser_submit.set_defaults(_func=submit_function)

    if merge_function is None:
        return subparser_main, subparser_submit
    else:
        subparser_merge = subparsers.add_parser('merge', parents=merge_parents)
        subparser_merge.set_defaults(_func=merge_function)
        return subparser_main, subparser_submit, subparser_merge

# see add_main_and_submit
def choose_parser_and_run(parser):
    args, _ = parser.parse_known_args()
    pretty.print_namespace(args); print()

    args._func(args)


